#include <drm/drm_pci.h>
#include "inspur_drm_drv.h"
#include "inspur_drm_regs.h"

void colorcur2monocur(void *data, void *out)
{
    unsigned int *col = (unsigned int *)data;
    unsigned char *mono = (unsigned char *)out;
    unsigned char pixel = 0;
    char bit_values;

    int i;
    for (i = 0; i < 64 * 64; i++)
    {
        if (*col >> 24 < 0xe0)
        {
            bit_values = 0;
        }
        else
        {
            int val = *col & 0xff;

            if (val < 0x80)
            {
                bit_values = 1;
            }
            else
            {
                bit_values = 2;
            }
        }
        col++;
        /* Copy bits into cursor byte */
        switch (i & 3)
        {
        case 0:
            pixel = bit_values;
            break;

        case 1:
            pixel |= bit_values << 2;
            break;

        case 2:
            pixel |= bit_values << 4;
            break;

        case 3:
            pixel |= bit_values << 6;
            *mono = pixel;
            mono++;
            pixel = 0;
            break;
        }
    }
}

static void inspur_hide_cursor(struct inspur_drm_private *priv)
{
    unsigned int reg = readl(priv->mmio + INSPUR_HWC_ADDRESS);
    reg &= ~INSPUR_HWC_ADDRESS_ENABLE_MASK;
    writel(reg, priv->mmio + INSPUR_HWC_ADDRESS);
}

static int inspur_show_cursor(struct inspur_drm_private *priv, void *src, unsigned int width, unsigned int height)
{
    struct drm_device *dev = priv->dev;
    struct drm_gem_vram_object *gbo;
    void *dst;
    s64 off;
    int ret;

    gbo = priv->cursor.gbo[priv->cursor.next_index];
    if (!gbo)
    {
        inspur_hide_cursor(priv);
        return -ENOTSUPP;
    }
    dst = drm_gem_vram_kmap(gbo, true, NULL);
    if (IS_ERR(dst)) {
        ret = PTR_ERR(dst);
        dev_err(&dev->pdev->dev, "failed to map cursor updates: %d\n", ret);
        return ret;
    }
    off = drm_gem_vram_offset(gbo);
    if (off < 0) {
        ret = (int)off;
        dev_err(&dev->pdev->dev, "failed to get cursor scanout address: %d\n", ret);
        goto err_drm_gem_vram_vunmap;
    }

    colorcur2monocur(src, dst);

    writel(0xffff0000, priv->mmio + INSPUR_HWC_COLOR_12);
    writel(0xffff, priv->mmio + INSPUR_HWC_COLOR_3);
    writel(INSPUR_FIELD(INSPUR_HWC_ADDRESS_ENABLE, 1) | INSPUR_FIELD(INSPUR_HWC_ADDRESS_ADDRESS, off),
           priv->mmio + INSPUR_HWC_ADDRESS);

    drm_gem_vram_kunmap(gbo);
    ++priv->cursor.next_index;
    priv->cursor.next_index %= ARRAY_SIZE(priv->cursor.gbo);
    return 0;

err_drm_gem_vram_vunmap:
    drm_gem_vram_kunmap(gbo);
    return ret;
}

int inspur_crtc_cursor_set(
            struct drm_crtc *crtc,
            struct drm_file *file_priv,
            uint32_t handle,
            uint32_t width,
            uint32_t height
)
{
    struct drm_device *dev = crtc->dev;
    struct inspur_drm_private *priv = (struct inspur_drm_private *)dev->dev_private;
    struct drm_gem_object *obj;
	struct drm_gem_vram_object *gbo = NULL;
    int ret;
    u8 *src;

    if (!handle || !file_priv) {
        inspur_hide_cursor(priv);
        return 0;
    }

    if ((width > 64 || height > 64) && handle) {
        unsigned int reg = readl(priv->mmio + INSPUR_HWC_ADDRESS);
        reg &= ~INSPUR_HWC_ADDRESS_ENABLE_MASK;
        writel(reg, priv->mmio + INSPUR_HWC_ADDRESS);
        return -EINVAL;
    }

    obj = drm_gem_object_lookup(file_priv, handle);
    if (!obj)
        return -ENOENT;
    gbo = drm_gem_vram_of_gem(obj);
    src = drm_gem_vram_kmap(gbo, true, NULL);
    if (IS_ERR(src)) {
        ret = PTR_ERR(src);
        dev_err(&dev->pdev->dev, "failed to map user buffer updates\n");
        goto err_drm_gem_object_put_unlocked;
    }

    ret = inspur_show_cursor(priv, src, width, height);
    if (ret) {
        goto err_drm_gem_vram_vunmap;
    }

    drm_gem_vram_kunmap(gbo);
    drm_gem_object_put_unlocked(obj);

    return 0;

err_drm_gem_vram_vunmap:
    drm_gem_vram_kunmap(gbo);
err_drm_gem_object_put_unlocked:
    drm_gem_object_put_unlocked(obj);
    return ret;
}

int inspur_cursor_init(struct inspur_drm_private *priv)
{
    struct drm_device *dev = priv->dev;
    int ret;
    int i = 0;
    struct drm_gem_vram_object *gbo;

    for (i = 0; i < ARRAY_SIZE(priv->cursor.gbo); i++)
    {
        gbo = drm_gem_vram_create(dev, &dev->vram_mm->bdev, roundup(1024, PAGE_SIZE), 0, false);
        if (IS_ERR(gbo)) {
            ret = PTR_ERR(gbo);
            goto err_drm_gem_vram_put;
        }
        ret = drm_gem_vram_pin(gbo, 0);
        if (ret) {
            drm_gem_vram_put(gbo);
            goto err_drm_gem_vram_put;
        }
        priv->cursor.gbo[i] = gbo;
    }
    return 0;
err_drm_gem_vram_put:
    while (i) {
        --i;
        gbo = priv->cursor.gbo[i];
        drm_gem_vram_unpin(gbo);
        drm_gem_vram_put(gbo);
        priv->cursor.gbo[i] = NULL;
    }
    return ret;
}

void inspur_cursor_fini(struct inspur_drm_private *priv)
{
    int i;
    struct drm_gem_vram_object *gbo;
    for (i = 0; i < ARRAY_SIZE(priv->cursor.gbo); ++i) {
        gbo = priv->cursor.gbo[i];
        drm_gem_vram_unpin(gbo);
        drm_gem_vram_put(gbo);
    }
}

int inspur_crtc_cursor_move(struct drm_crtc *crtc, int x, int y)
{
    struct inspur_drm_private *priv __attribute__((unused)) = (struct inspur_drm_private *)crtc->dev->dev_private;

    DRM_DEBUG_KMS("(%d, %d)\n", x, y);
    writel(INSPUR_FIELD(INSPUR_HWC_LOCATION_X, x) | INSPUR_FIELD(INSPUR_HWC_LOCATION_Y, y),
        priv->mmio + INSPUR_HWC_LOCATION);

    return 0;
}

#define HW_FLAG_OFFSET 0x01ffff00
#define HW_FLAG_ENABLE 0x1bd40750
unsigned char getKVMHWCursorSetting(struct inspur_drm_private *priv)
{
    unsigned int value = *(unsigned int*)(priv->fb_map + HW_FLAG_OFFSET);
    DRM_DEBUG_KMS("HW_FLAG = %x\n", value);
#if 0
    if (value == HW_FLAG_ENABLE)
        return 1;
    else
#endif
        return 0;
}
