/* INSPUR SoC drm driver
 *
 * Based on the smi drm driver.
 *
 * Copyright (c) 2020 SMI Limited.
 *
 * Author:
 *	
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 */

#include <drm/drm_atomic_helper.h>

#include "inspur_drm_drv.h"

int inspur_mm_init(struct inspur_drm_private *inspur)
{
	struct drm_vram_mm *vmm;
	int ret;
	struct drm_device *dev = inspur->dev;

	vmm = drm_vram_helper_alloc_mm(dev,
				       pci_resource_start(dev->pdev, 0),
				       inspur->fb_size, &drm_gem_vram_mm_funcs);
	if (IS_ERR(vmm)) {
		ret = PTR_ERR(vmm);
		DRM_ERROR("Error initializing VRAM MM; %d\n", ret);
		return ret;
	}

	return 0;
}

void inspur_mm_fini(struct inspur_drm_private *inspur)
{
	if (!inspur->dev->vram_mm)
		return;

	drm_vram_helper_release_mm(inspur->dev);
}

int inspur_gem_create(struct drm_device *dev, u32 size, bool iskernel,
		     struct drm_gem_object **obj)
{
	struct drm_gem_vram_object *gbo;
	int ret;

	*obj = NULL;

	size = roundup(size, PAGE_SIZE);
	if (size == 0)
		return -EINVAL;

	gbo = drm_gem_vram_create(dev, &dev->vram_mm->bdev, size, 0, false);
	if (IS_ERR(gbo)) {
		ret = PTR_ERR(gbo);
		if (ret != -ERESTARTSYS)
			DRM_ERROR("failed to allocate GEM object: %d\n", ret);
		return ret;
	}
	*obj = &gbo->bo.base;
	return 0;
}

int inspur_dumb_create(struct drm_file *file, struct drm_device *dev,
		      struct drm_mode_create_dumb *args)
{
	struct drm_gem_object *gobj;
	u32 handle;
	int ret;

	args->pitch = ALIGN(args->width * DIV_ROUND_UP(args->bpp, 8), 128);
	args->size = args->pitch * args->height;

	ret = inspur_gem_create(dev, args->size, false, &gobj);
	if (ret) {
		DRM_ERROR("failed to create GEM object: %d\n", ret);
		return ret;
	}

	ret = drm_gem_handle_create(file, gobj, &handle);
	drm_gem_object_put_unlocked(gobj);
	if (ret) {
		DRM_ERROR("failed to unreference GEM object: %d\n", ret);
		return ret;
	}

	args->handle = handle;
	return 0;
}


static void inspur_user_framebuffer_destroy(struct drm_framebuffer *fb)
{
	drm_gem_object_put_unlocked(fb->obj[0]);
	drm_framebuffer_cleanup(fb);
	kfree(fb);
}

static const struct drm_framebuffer_funcs inspur_fb_funcs = {
	.destroy = inspur_user_framebuffer_destroy,
};

struct inspur_framebuffer *
inspur_framebuffer_init(struct drm_device *dev,
		       const struct drm_mode_fb_cmd2 *mode_cmd,
		       struct drm_gem_object *obj)
{
	struct inspur_framebuffer *inspur_fb;
	int ret;

	inspur_fb = kzalloc(sizeof(*inspur_fb), GFP_KERNEL);
	if (!inspur_fb) {
		DRM_ERROR("failed to allocate inspur_fb\n");
		return ERR_PTR(-ENOMEM);
	}

	drm_helper_mode_fill_fb_struct(dev, inspur_fb, mode_cmd);
	inspur_fb->obj[0] = obj;
	ret = drm_framebuffer_init(dev, inspur_fb, &inspur_fb_funcs);
	if (ret) {
		DRM_ERROR("drm_framebuffer_init failed: %d\n", ret);
		kfree(inspur_fb);
		return ERR_PTR(ret);
	}

	return inspur_fb;
}

static struct drm_framebuffer *
inspur_user_framebuffer_create(struct drm_device *dev,
			      struct drm_file *filp,
			      const struct drm_mode_fb_cmd2 *mode_cmd)
{
	struct drm_gem_object *obj;
	struct inspur_framebuffer *inspur_fb;

	DRM_DEBUG_DRIVER("%dx%d, format %c%c%c%c\n",
			 mode_cmd->width, mode_cmd->height,
			 (mode_cmd->pixel_format) & 0xff,
			 (mode_cmd->pixel_format >> 8)  & 0xff,
			 (mode_cmd->pixel_format >> 16) & 0xff,
			 (mode_cmd->pixel_format >> 24) & 0xff);

	obj = drm_gem_object_lookup(filp, mode_cmd->handles[0]);
	if (!obj)
		return ERR_PTR(-ENOENT);

	inspur_fb = inspur_framebuffer_init(dev, mode_cmd, obj);
	if (IS_ERR(inspur_fb)) {
		drm_gem_object_put_unlocked(obj);
		return ERR_PTR((long)inspur_fb);
	}
	return inspur_fb;
}

const struct drm_mode_config_funcs inspur_mode_funcs = {
	.atomic_check = drm_atomic_helper_check,
	.atomic_commit = drm_atomic_helper_commit,
	.fb_create = inspur_user_framebuffer_create,
};
